from pprint import pprint
from timemachines.optimizers.alloptimizers import OPTIMIZERS
from timemachines.optimizers.objectives import OBJECTIVES
import random
from typing import List
import numpy as np
from pprint import pprint
from ratings.elo import elo_expected


def optimizer_name(solver):
    return solver.__name__.replace('_cube','')


def objective_name(objective):
    return objective.__name__.replace('_cube','')


def debug():
    for n_trials in [2,10,40]:
        for n_dim in [2,3]:
            for optimizer in OPTIMIZERS:
                for objective in OBJECTIVES:
                    try:
                        optimizer(objective, n_trials=n_trials,n_dim=n_dim,with_count=True)
                    except Exception as e:
                        print(e)
                        print(optimizer_name(optimizer) + ' fails on ' + objective.__name__ +
                              ' in dimension '+str(n_dim)+' when n_trials='+str(n_trials))


def comparison():
    n_trials,n_dim = 100, 5
    comparison = sorted([(objective.__name__,optimizer(objective, n_trials=n_trials, n_dim=n_dim, with_count=True),
                          optimizer_name(optimizer))
                         for optimizer in OPTIMIZERS for objective in OBJECTIVES])
    pprint(comparison)


def random_matchup(optimizers:List, objectives:List) -> dict:
    optim1, optim2 = np.random.choice( optimizers, 2, replace=False )
    objective = np.random.choice( objectives, 1 )[0]
    return matchup(optim1,optim2,objective)


def matchup( white, black, objective )-> dict:
    """ Compare two optimizers, making an effort to equalize function evaluations
    :param white:   optimizer
    :param black:   optimizer
    :param objective:
    :return: match result dictionary
    """
    n_trials = np.random.choice( [50,100,200,400,800,1600] )
    n_dim = np.random.choice( [1,2,4,6,8,10,12,16] )
    best_val_white, best_x_white, count_white = white(objective, n_trials=n_trials, n_dim=n_dim, with_count=True)
    best_val_black, best_x_black, count_black = black(objective, n_trials=int(0.8*count_white),  n_dim=n_dim, with_count=True)
    valid_white, valid_black = count_white<2*n_trials, count_black <= n_trials

    tol = 1e-3*(abs(best_val_white)+abs(best_val_black))
    if valid_white and valid_black:
        if best_val_white < best_val_black - tol:
            points = 1.0
        elif best_val_black < best_val_white - tol:
            points = 0.0
        else:
            points = 0.5
    else:
        points = None

    return {'white':optimizer_name(white),'black':optimizer_name(black),'points':points,'n_dim':n_dim,
            'n_trials':n_trials,'white_best_val':best_val_white,'black_best_val':best_val_black,
            'white_trails':count_white,'black_trials':count_black}




if __name__=='__main__':
    if True:
        comparison()
    if False:
        res = random_matchup(optimizers=OPTIMIZERS,objectives=OBJECTIVES)
        pprint(res)






from timemachines.optimizers.hyperoptcube import HYPEROPT_OPTIMIZERS
from timemachines.optimizers.shgocube import SHGO_OPTIMIZERS
from timemachines.optimizers.optunacube import OPTUNA_OPTIMIZERS
from timemachines.optimizers.pysotcube import PYSOT_OPTIMIZERS
from timemachines.optimizers.scipycube import SCIPY_OPTIMIZERS
from timemachines.optimizers.axcube import AX_OPTIMIZERS
from timemachines.optimizers.platypuscube import PLATYPUS_OPTIMIZERS
from timemachines.optimizers.pymoocube import PYMOO_OPTMIZERS

CANDIDATES = SCIPY_OPTIMIZERS + SHGO_OPTIMIZERS + HYPEROPT_OPTIMIZERS +\
             PYSOT_OPTIMIZERS + OPTUNA_OPTIMIZERS + AX_OPTIMIZERS +\
             PLATYPUS_OPTIMIZERS + PYMOO_OPTMIZERS

# To see what might be working, or not, refer to directories such as:
# https://github.com/microprediction/timemachines-testing/tree/main/data/brownian/dlm_seasonal
OPTIMIZERS = SCIPY_OPTIMIZERS + PYSOT_OPTIMIZERS + AX_OPTIMIZERS + OPTUNA_OPTIMIZERS



if __name__=='__main__':
    print(str(len(OPTIMIZERS)) + ' in use')
    from timemachines.optimizers.objectives import OBJECTIVES
    for objective in OBJECTIVES:
        print(' ')
        print(objective.__name__)
        for optimizer in OPTIMIZERS:
            print(optimizer.__name__,(optimizer.__name__,optimizer(objective, n_trials=50, n_dim=5, with_count=True)))



from timemachines.optimization import optimize
from timemachines.synthetic import brownian_with_exogenous
from timemachines.conventions import from_space
from timemachines.optimizers.alloptimizers import OPTIMIZERS
from timemachines.skaters.pmd import pmd_hyperparams, pmd_auto
from pprint import pprint
import random


def optimize_dlm():
    optimizer = random.choice(OPTIMIZERS)
    print('Running '+str(optimizer.__name__))
    best_val, best_x = optimize(f=pmd_auto, ys=brownian_with_exogenous(n=300),
                      n_trials=25, n_dim=3, optimizer=optimizer)
    best_r = from_space(best_x)
    s = pmd_hyperparams(s=dict(), r=best_r)
    print("Best hyper-param is " + str(best_r)+' with interpretation ')
    pprint(s)


from timemachines.skaters.pmd import pmd_auto
from timemachines.skaters.dlmseasonal import dlm_seasonal
from timemachines.skaters.dlmauto import dlm_auto

SKATERS = [ pmd_auto, dlm_auto, dlm_seasonal ]


from pydlm import dlm, trend, seasonality, autoReg
from timemachines.conventions import dimension, to_int_log_space, separate_observations, update_buffers, \
    Y_TYPE, A_TYPE, R_TYPE
from timemachines.plotting import prior_plot_exogenous
import math
import numpy as np
import matplotlib.pyplot as plt


class fixedAutoReg(autoReg):

    def __init__(self,*arg,**kwargs):
        super().__init__(*arg,**kwargs)

    def appendNewData(self, data):
        """ AutoReg append new data automatically with the main time series. Nothing
        needs to be done here.
        """
        return


def dlm_auto_hyperparams(s, r:R_TYPE):
    # Seasonal model
    period_choices = [3,5,7,10,12,16,24,32]
    s['auto_degree'], s['trend_degree'], period_choice = to_int_log_space(r,bounds=[
        (1,10),   # auto-reg
        (0,3),    # seasonality degree
        (0,len(period_choices)) # seasonality period
    ])
    s['discount'] = 1 - ( r % 0.1 ) # Shouldn't matter too much
    s['period'] = period_choices[period_choice]
    s['n_burn'] = 100
    s['n_fit']  = 500  # How often to tune discount subsequently
    return s


def dlm_auto_or_last_value(s, k:int, y:Y_TYPE):
    num_obs = len(s['model'].data) if s.get('model') else 0
    if num_obs < s['n_burn']:
        y0, _ = separate_observations(y,s['dim'])
        return y0, s, None
    else:
        assert k==1,'only k==1 for now' # TODO: Fix
        x_mean, x_var = s['model'].predict()
        x = x_mean[0, 0]
        v = x_var[0,0]
        try:
            w = math.sqrt(v)
        except:
            print('Warning - negative var returned '+str(v))
            w = math.sqrt(abs(v))
        return x, s, w


def dlm_auto(y, s, k, a, t, e, r):
    """ One way to use dlm
        :returns: x, s', w
    """
    if s is None:
        s = dict()
        s = dlm_auto_hyperparams(s=s, r=r)
        s['dim'] = dimension(y)
        s['n_obs'] = 0
        s['model'] = dlm([],printInfo=False) + trend(s['trend_degree'], s['discount']) + seasonality(s['period'], s['discount'])
        s['model'] = s['model'] + fixedAutoReg(degree=s['auto_degree'], name='ar', w=1.0)

    if y is not None:
        s['n_obs'] += 1
        assert isinstance(y, float) or len(y) == s['dim'], ' Cannot change dimension of input in flight '
        y0, exog = separate_observations(y=y,dim=s['dim'])
        y0_passed_in = None if np.isnan(y0) else None  # pydlm uses None for missing values
        s['model'].append([y0])
        num_obs = len(s['model'].data) if s.get('model') else 0
        if num_obs % s['n_fit'] == s['n_fit']-1:
            _, s, _ = dlm_auto(y=None,s=s,k=k,a=a,t=t,e=10,r=r)
        s['model'].fitForwardFilter()
        s['model'].fitBackwardSmoother()
        return dlm_auto_or_last_value(s=s, k=k, y=y)

    if y is None:
        s['model'].tune() # Tunes discount factors
        s['model'].fit()
        return None, s, None


if __name__ == '__main__':
    err = prior_plot_exogenous(f=dlm_auto, k=1, n=1000, r=np.random.rand())
    plt.figure()
    print('done')
    pass
